
open Ast
open Sast


let string_of_op = function
		Add -> "+"
	| Sub -> "-"
  | Mult -> "*"
  | Div -> "/"
  | Series -> "#"
  | Parallel -> "$"
  | Eq -> "=="
  | Ne -> "!="
  | Gt -> ">"
  | Lt -> "<"
  | Ge -> ">="
  | Le -> "<="
  
let rec string_of_obj_type t = match t with
    Int -> "int"
  | Float -> "float"
  | String -> "string"
  | Bit -> "bit"
  | Res -> "res"
  | Void -> "void"
  | Boolean -> "boolean"
  
type symbol_table = {
  parent : symbol_table option;
  variables : Sast.var_decl list;
  functions: Sast.func_decl list;
}

type trans_env = {
  scope : symbol_table;
}

let rec find_variable (scope : symbol_table) name =
  try
    List.find (fun v -> v.vvname = name) scope.variables
  with Not_found ->
    match scope.parent with
    Some(parent) -> find_variable parent name
    | _ -> raise (Failure("variable " ^ name ^ " not defined"))
    
let var_exists scope name =
    try
      let _ = find_variable scope name
      in true
    with Failure(_) ->
      false

let rec find_function (scope : symbol_table) name = match name with
	"show" ->  {frtype = Void; ffname = "show";fformals=[];flocals=[];fbody=[] }
  | _ -> (try
    List.find (fun f -> f.ffname = name) scope.functions
  with Not_found ->
    match scope.parent with
    Some(parent) -> find_function parent name
    | _ -> raise (Failure("function " ^ name ^ " not defined")))


let func_exists scope name =
    List.exists (fun f -> f.ffname = name) scope.functions

let assign_allowed lt rt = match lt with
      Res ->  (rt = Int) || (rt = Float) || (rt = Res)  
    | Bit ->  (rt = Bit) || (rt = String)
    | _ -> lt = rt

let rec can_assign lt rval = 
  let (_, rt) = rval in
    if assign_allowed lt rt then
      rval
    else
      raise (Failure("type " ^ string_of_obj_type rt ^ " cannot be put into type " ^ string_of_obj_type lt))

let to_ast ast_vdecl sast_vdecl = 
  {vtype = sast_vdecl.vvtype; vname = sast_vdecl.vvname}


let can_op lval op rval = 
  let (_, lt) = lval
  and (_, rt) = rval in                   
  let type_match = (lt = rt) in
  let int_match = (lt = Int) in
  let float_match = (lt = Float) in
  let string_match = (lt = String) in
  let bit_match = (lt = Bit) in
  let res_match = (lt = Res) in
  let boolean_match = (lt = Boolean) in
  let result = match op with
      Ast.Add     -> (type_match && (int_match || float_match || string_match || bit_match || res_match )), lt
    | Ast.Sub     -> (type_match && (int_match || float_match || bit_match || res_match ) ), lt
    | Ast.Mult    -> (type_match && (int_match || float_match || bit_match || res_match ) ), lt
    | Ast.Div     -> (type_match && (int_match || float_match || bit_match || res_match ) ), lt
    | Ast.Series  -> (type_match && res_match ), lt
    | Ast.Parallel  -> (type_match && res_match), lt
    | Ast.Eq -> (type_match && (int_match || float_match || bit_match || res_match )), Boolean
    | Ast.Ne -> (type_match && (int_match || float_match || bit_match || res_match)), Boolean
    | Ast.Gt -> (type_match && (int_match || float_match || bit_match || res_match)), Boolean
    | Ast.Lt -> (type_match && (int_match || float_match || bit_match || res_match)), Boolean
    | Ast.Ge -> (type_match && (int_match || float_match || bit_match || res_match)), Boolean
    | Ast.Le -> (type_match && (int_match || float_match || bit_match || res_match)), Boolean

  in if fst result then
    snd result
  else
    raise (Failure("operator" ^ string_of_op op ^ " cannot be used on types " ^
      string_of_obj_type lt ^ " and " ^ string_of_obj_type rt))

let translate (globals, funcs) =
  let rec trans_expr env = function
      Ast.Id(n) -> let vdecl = (find_variable env.scope n) in
                     Sast.Id(vdecl), vdecl.vvtype
     | Ast.Unary (un_op, e) ->
						let et = trans_expr env e in
							let _,t = et in            (*expr and type*)
							  let tt = match un_op with
								  | Reciprocal -> if t = Int or t = Float then t else
										(*TODO: t!=0*)
											raise (Failure ("Only integers and floats are allowed for using reciprocal"))
								(*  | Not -> if t = Boolean then Boolean else
											raise (Failure ("Only boolean is allowed for boolean operators"))  *)
				          | Plus -> if t = Int or t = Float then t else
											raise (Failure ("Only integers and floats can be added a positive sign"))
				          | Minus -> if t = Int or t = Float then t else
											raise (Failure ("Only integers and floats can be added a negative sign"))
	 			          | _ -> raise (Failure ("The operator is not unary"))
						in
						Sast.Unary(un_op, et), tt
		 | Ast.Afterop (e, after_op) ->
		        let et = trans_expr env e in
		        let _,t = et in
						  let tt = match after_op with
							  | Dot -> if t = Res or t = Bit then t else
									(*TODO: t!=0*)
										raise (Failure ("Only res or bit is allowed for dot value."))
							  | _ -> raise (Failure ("The operator is not afterop."))
						in
						Sast.Afterop (et, after_op), tt
  		| Ast.Literal(lit) -> (match lit with
					| IntLit i -> Literal(lit), Int
					| FloatLit f -> Literal(lit), Float
					| BinaryLit b -> Literal(lit), Bit
					| SentenceLit s -> Literal(lit), String
	      )
	    | Ast.Binop(e1, op, e2) ->
	        let e1 = trans_expr env e1
	        and e2 = trans_expr env e2 
	        in let rtype = can_op e1 op e2 in
	        Sast.Binop(e1, op, e2), rtype
	
	   | Ast.Call (func_name, params) ->
	       let func_decl_call = 
	       (find_function env.scope func_name)
	       in
	       let checked_fname = if (func_decl_call.ffname=func_name) then func_name
	          else raise (Failure("undeclaration of " ^ func_name))
	       in
	       let typed_params = List.map (trans_expr env) params 
	       in
				 Sast.Call(checked_fname, typed_params), func_decl_call.frtype

	    | Ast.Assign(lv, e) ->
	        let lval, t = (trans_expr env lv) in
	        let aval = (trans_expr env e) in
	        Sast.Assign((lval, t), (can_assign t aval)), t
	    | Ast.Noexpr ->
	        Sast.Noexpr, Void

	 in let add_local env v = 
		    let evalue = if (var_exists env.scope v.vname) = true then 
		        raise (Failure("redeclaration of " ^ v.vname))
		    in (**)
		    let new_v = {
		            vvname = v.vname;
		            vvtype = v.vtype;
		          
		      }
		    in let vars = new_v :: env.scope.variables
		    in let scope' = {env.scope with variables = vars}
		    in {(*env with*) scope = scope'}

   in let rec trans_stmt env = function


        Ast.Block(v, s) -> 
          let scope' = {parent = Some(env.scope); variables = []; functions = []}
	        in let env' = {(*env with*) scope = scope'}
	        in let block_env = List.fold_left add_local env' (List.rev v)
	        
	        in let s' = List.map (fun s -> trans_stmt block_env s) s
	        in let fvlist = block_env.scope.variables
	        in Sast.Block(fvlist, s')  

	    | Ast.Expr(e) ->
	        Sast.Expr(trans_expr env e)
	    | Ast.Return(e) ->
	        Sast.Return(trans_expr env e)
	    | Ast.If (e, s1, s2) ->
	        let e' = trans_expr env e
	        in Sast.If(can_assign Boolean e', trans_stmt env s1, trans_stmt env s2)
	    | Ast.While (e, s) ->
	        let e' = trans_expr env e
	        in Sast.While(can_assign Boolean e', trans_stmt env s)
	    | Ast.For (e1, e2, e3, s) ->
	        let e2' = trans_expr env e2
	        in Sast.For(trans_expr env e1, can_assign Boolean e2', trans_expr env e3, trans_stmt env s)

   in let add_func env f =
	    let new_f = match ((var_exists env.scope f.fname) || (func_exists env.scope f.fname)) with
	      true -> raise (Failure("redeclaration of " ^ f.fname))
	    | false ->
	        let scope' = {parent = Some(env.scope); variables = []; functions = []}
	        in let env' = {(*env with*) scope = scope'}
	        in let env' = List.fold_left add_local env' (List.rev f.formals)
	        in {
	          frtype = f.rtype;
	          ffname = f.fname;
	          fformals = env'.scope.variables;
	          flocals = [];
	          fbody = [];
	        }
	    in let funcs = new_f :: env.scope.functions
	    in let scope' = {env.scope with functions = funcs}
	    in {(*env with*) scope = scope'}  

	 in let trans_func env (f : Ast.func_decl) =  
	    let sf = find_function env.scope (f.fname)
	    in let functions' = List.filter (fun f -> f.ffname != sf.ffname) env.scope.functions
	    in let scope' = {parent = Some(env.scope); variables = sf.fformals; functions = []}
	    in let env' = {(*env with*) scope = scope'}
	    in let formals' = env'.scope.variables
	    in let env' = List.fold_left add_local env' (f.locals)
	    in let remove v =
	      not (List.exists (fun fv -> fv.vvname = v.vvname) formals')
	    in let locals' = List.filter remove env'.scope.variables 
	    in let body' = List.map (fun f -> trans_stmt env' f) (f.body)
	    in let new_f = {
	      sf with 
	      fformals = formals';
	      flocals = locals';
	      fbody = body';
	    }
	    in let funcs = new_f :: functions'
	    in let scope' = {env.scope with functions = funcs}
	    in {(*env with*) scope = scope'}

	 in let validate_func f =
	    let is_return = function
	        Sast.Return(e) -> true
	      | _ -> false
	    in let valid_return = function
	        Sast.Return(e) -> if assign_allowed f.frtype (snd e) then
	                            true
	                          else
	                             raise (Failure(  f.ffname ^ " must return type " ^
	                              string_of_obj_type f.frtype ^
	                              ", not " ^ string_of_obj_type (snd e)
	                             ))
	      | _ -> false
	    in let returns = List.filter is_return f.fbody
	    in let _ = List.for_all valid_return returns
	    in let return_count = List.length returns
	    in if (return_count = 0 && f.frtype != Void) then
	      raise (Failure(f.ffname ^ " must have a return type of " ^ string_of_obj_type f.frtype))
	    else if List.length f.fformals > 8 then
	      raise (Failure(f.ffname ^ " must have less than 8 formals"))
	    else
	      f                              

   in let global_scope = {
		    parent = None;
		    variables = [];
		    functions = [];
		  }
	 in let genv = {
		    scope = global_scope;
		  }
		  in let genv = List.fold_left add_local genv (List.rev globals)
		  in let genv = List.fold_left add_func genv (List.rev funcs)
		  in let genv = List.fold_left trans_func genv (List.rev funcs)
		  in if func_exists genv.scope "main" then
		    (genv.scope.variables, List.map validate_func genv.scope.functions)
		  else
		    raise (Failure("no main function defined"))
 